/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * whls - Weighted HLS <-> RGB conversions
 */

#include "whls.xh"

typedef struct {
      
      // rgb multipliers
      double rc, gc, bc;

      // how much light is given to grey when whls.value = 0.5
      double greyc;

      // gamma adjustment, 1.00 = no gamma
      double gamma;

} whls_profile_t;

const whls_profile_t default_prof = {
   rc: 0.37,
   gc: 0.43,
   bc: 0.20,
   greyc: 0.5,
   gamma: 0.59,
};

void rgb_to_whls(GdkColor *rgb, WHLSColor *whls) {
   double r = rgb->red / 65535.0;
   double g = rgb->green / 65535.0;
   double b = rgb->blue / 65535.0;

   double h, v, v2, vmin;

   if(r >= g) {
      if(r >= b) {
	 if(g >= b) {
	    if(r == b) {
	       v = vmin = r;
	       h = 0;
	    } else {
	       v = r; v2 = g; vmin = b;
	       h = (v2 - vmin) / (v - vmin);
	    }
	 } else {
	    v = r; v2 = b; vmin = g;
	    h = 6 - (v2 - vmin) / (v - vmin);
	 }
      } else {
	 v = b; v2 = r; vmin = g;
	 h = 4 + (v2 - vmin) / (v - vmin);
      }
   } else if(g >= b) {
      if(r >= b) {
	 v = g; v2 = r; vmin = b;
	 h = 2 - (v2 - vmin) / (v - vmin);
      } else {
	 v = g; v2 = b; vmin = r;
	 h = 2 + (v2 - vmin) / (v - vmin);
      }
   } else {
      v = b; v2 = g; vmin = r;
      h = 4 - (v2 - vmin) / (v - vmin);
   }

   whls->h = h;

   double s = v > 0 ? 1 - vmin / v : 0;
   /*
     double imv2 = 2 + s - 2 * v;
     ms = imv2 > 0 ? imv2 >= 1 ? s : s / imv2 : 0;
   */
   double imv2 = v - s * (1 - default_prof.greyc);

   whls->ms = imv2 == 1 ? 0 : imv2 <= default_prof.greyc ? s : s * (1 - default_prof.greyc) / (1 - imv2);

   whls->wv = r * default_prof.rc + g * default_prof.gc + b * default_prof.bc;
}

static double hf(double v, int c) {
   // v *= 6;
   v -= c;
   while(v < -3) v += 6;
   while(3 < v) v -= 6;
   if(-1 <= v && v <= 1) return 1;
   if(-2 < v && v < -1) return v+2;
   if(1 < v && v < 2) return 2-v;
   return 0;
}

static double chr(double v) {
   return hf(v, 0);
}

static double chg(double v) {
   return hf(v, 2);
}

static double chb(double v) {
   return hf(v, 4);
}

void whls_to_rgb(WHLSColor *whls, GdkColor *rgb) {
    double h = whls->h;
    double ms = whls->ms;
    double wv = whls->wv;

    //  wv = Math.pow(wv, 1/default_prof.gamma);

    double hr = chr(h);
    double hg = chg(h);
    double hb = chb(h);

    // hr = Math.pow(hr, 1/default_prof.gamma); hg = Math.pow(hg, 1/default_prof.gamma); hb = Math.pow(hb, 1/default_prof.gamma);

    double whsum = hr * default_prof.rc + hg * default_prof.gc + hb * default_prof.bc;
    double s, v;

    if(ms == 0) {
      // middle axis, greyscale
      s = 0;
      v = wv;
    /*
    } else if(wv > whsum) {
      // upper cone
      System.out.print("^ ");
      double t = ms * (1 - wv);
      s = t / (1 - whsum);
      v = wv + t;
    } else {
      System.out.print("v ");
      // lower cone or grey
      s = ms;
      v = s * wv / whsum + (1 - s) * wv;
    }
    */
      //    } else if(wv * ms / whsum + (wv * (1 - ms)) / 0.5 > 1) { //  || (ms >= 0.399 && ms <= 0.401)) {
    } else if(wv / (whsum * ms + (1 - ms) * (1-whsum)) > 1) { //  || (ms >= 0.399 && ms <= 0.401)) {
      // upper cone
      double t = ms * (1 - wv);
      s = t / (1 - whsum);
      //v = wv + t;
      v = wv / ( 1 - ms * (1 - wv));
    } else {
      // lower cone or grey
//      v = s * wv / whsum + (1 - s) * wv;
//      v = s * wv / whsum;
      v = wv / ( whsum + (1 - whsum) * (1 - ms));
      //s = ms == 1 ? 1 : (2 * ms * (1 - v)) / (1 - ms);
      s = ms;
    }

    double spart = s * v;
    double gpart = (1 - s) * v;

    rgb->red   = (hr * spart + gpart) * 65535.0;
    rgb->green = (hg * spart + gpart) * 65535.0;
    rgb->blue  = (hb * spart + gpart) * 65535.0;
    
    // r = Math.pow(r, default_prof.gamma); g = Math.pow(g, default_prof.gamma); b = Math.pow(b, default_prof.gamma);
}
